# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import copy

import awkward as ak
from awkward._v2.index import Index
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.indexedform import IndexedForm
from awkward._v2.forms.form import _parameters_equal

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


class IndexedArray(Content):
    is_IndexedType = True

    def __init__(self, index, content, identifier=None, parameters=None, nplike=None):
        if not (
            isinstance(index, Index)
            and index.dtype
            in (
                np.dtype(np.int32),
                np.dtype(np.uint32),
                np.dtype(np.int64),
            )
        ):
            raise TypeError(
                "{} 'index' must be an Index with dtype in (int32, uint32, int64), "
                "not {}".format(type(self).__name__, repr(index))
            )
        if not isinstance(content, Content):
            raise TypeError(
                "{} 'content' must be a Content subtype, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if nplike is None:
            nplike = content.nplike
        if nplike is None:
            nplike = index.nplike

        self._index = index
        self._content = content
        self._init(identifier, parameters, nplike)

    @property
    def index(self):
        return self._index

    @property
    def content(self):
        return self._content

    Form = IndexedForm

    def _form_with_key(self, getkey):
        form_key = getkey(self)
        return self.Form(
            self._index.form,
            self._content._form_with_key(getkey),
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=form_key,
        )

    def _to_buffers(self, form, getkey, container, nplike):
        assert isinstance(form, self.Form)
        key = getkey(self, form, "index")
        container[key] = ak._v2._util.little_endian(self._index.raw(nplike))
        self._content._to_buffers(form.content, getkey, container, nplike)

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return IndexedArray(
            ak._v2.index.Index(self._index.raw(tt)),
            self._content.typetracer,
            self._typetracer_identifier(),
            self._parameters,
            tt,
        )

    @property
    def length(self):
        return self._index.length

    def _forget_length(self):
        return IndexedArray(
            self._index.forget_length(),
            self._content,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<IndexedArray len="]
        out.append(repr(str(self.length)))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._index._repr(indent + "    ", "<index>", "</index>\n"))
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</IndexedArray>")
        out.append(post)
        return "".join(out)

    def merge_parameters(self, parameters):
        return IndexedArray(
            self._index,
            self._content,
            self._identifier,
            ak._v2._util.merge_parameters(self._parameters, parameters),
            self._nplike,
        )

    def toIndexedOptionArray64(self):
        return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
            self._index, self._content, self._identifier, self._parameters, self._nplike
        )

    def mask_as_bool(self, valid_when=True):
        if valid_when:
            return self._index.data >= 0
        else:
            return self._index.data < 0

    def _getitem_nothing(self):
        return self._content._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if not self._nplike.known_data:
            return self._content._getitem_at(where)

        if where < 0:
            where += self.length
        if self._nplike.known_shape and not 0 <= where < self.length:
            raise NestedIndexError(self, where)
        return self._content._getitem_at(self._index[where])

    def _getitem_range(self, where):
        if not self._nplike.known_shape:
            return self

        start, stop, step = where.indices(self.length)
        assert step == 1
        return IndexedArray(
            self._index[start:stop],
            self._content,
            self._range_identifier(start, stop),
            self._parameters,
            self._nplike,
        )

    def _getitem_field(self, where, only_fields=()):
        return IndexedArray(
            self._index,
            self._content._getitem_field(where, only_fields),
            self._field_identifier(where),
            None,
            self._nplike,
        )

    def _getitem_fields(self, where, only_fields=()):
        return IndexedArray(
            self._index,
            self._content._getitem_fields(where, only_fields),
            self._fields_identifier(where),
            None,
            self._nplike,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        try:
            nextindex = self._index[carry.data]
        except IndexError as err:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, str(err))
            else:
                raise exception(str(err))

        return IndexedArray(
            nextindex,
            self._content,
            self._carry_identifier(carry, exception),
            self._parameters,
            self._nplike,
        )

    def _getitem_next_jagged_generic(self, slicestarts, slicestops, slicecontent, tail):
        if self._nplike.known_shape and slicestarts.length != self.length:
            raise NestedIndexError(
                self,
                ak._v2.contents.ListArray(
                    slicestarts, slicestops, slicecontent, None, None, self._nplike
                ),
                "cannot fit jagged slice with length {} into {} of size {}".format(
                    slicestarts.length, type(self).__name__, self.length
                ),
            )

        nextcarry = ak._v2.index.Index64.empty(self.length, self._nplike)
        assert nextcarry.nplike is self._nplike and self._index.nplike is self._nplike
        self._handle_error(
            self._nplike[
                "awkward_IndexedArray_getitem_nextcarry",
                nextcarry.dtype.type,
                self._index.dtype.type,
            ](
                nextcarry.data,
                self._index.data,
                self._index.length,
                self._content.length,
            )
        )
        # an eager carry (allow_lazy = false) to avoid infinite loop (unproven)
        next = self._content._carry(nextcarry, False, NestedIndexError)
        return next._getitem_next_jagged(slicestarts, slicestops, slicecontent, tail)

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        return self._getitem_next_jagged_generic(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, (int, slice, ak._v2.index.Index64)):
            nexthead, nexttail = ak._v2._slicing.headtail(tail)

            nextcarry = ak._v2.index.Index64.empty(self._index.length, self._nplike)
            assert (
                nextcarry.nplike is self._nplike and self._index.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_IndexedArray_getitem_nextcarry",
                    nextcarry.dtype.type,
                    self._index.dtype.type,
                ](
                    nextcarry.data,
                    self._index.data,
                    self._index.length,
                    self._content.length,
                )
            )

            next = self._content._carry(nextcarry, False, NestedIndexError)
            return next._getitem_next(head, tail, advanced)

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            raise NotImplementedError

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def project(self, mask=None):
        if mask is not None:
            if self._nplike.known_shape and self._index.length != mask.length:
                raise ValueError(
                    "mask length ({}) is not equal to {} length ({})".format(
                        mask.length(), type(self).__name__, self._index.length
                    )
                )
            nextindex = ak._v2.index.Index64.empty(self._index.length, self._nplike)
            assert (
                nextindex.nplike is self._nplike
                and mask.nplike is self._nplike
                and self._index.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_IndexedArray_overlay_mask",
                    nextindex.dtype.type,
                    mask.dtype.type,
                    self._index.dtype.type,
                ](
                    nextindex.data,
                    mask.data,
                    self._index.data,
                    self._index.length,
                )
            )
            next = ak._v2.contents.indexedoptionarray.IndexedOptionArray(
                nextindex,
                self._content,
                self._identifier,
                self._parameters,
                self._nplike,
            )
            return next.project()

        else:
            nextcarry = ak._v2.index.Index64.empty(self.length, self._nplike)
            assert (
                nextcarry.nplike is self._nplike and self._index.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_IndexedArray_getitem_nextcarry",
                    nextcarry.dtype.type,
                    self._index.dtype.type,
                ](
                    nextcarry.data,
                    self._index.data,
                    self._index.length,
                    self._content.length,
                )
            )
            return self._content._carry(nextcarry, False, NestedIndexError)

    def simplify_optiontype(self):
        if isinstance(
            self._content,
            (
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):

            if isinstance(
                self._content,
                (
                    ak._v2.contents.indexedarray.IndexedArray,
                    ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ),
            ):
                inner = self._content.index
                result = ak._v2.index.Index64.empty(self.index.length, self._nplike)
            elif isinstance(
                self._content,
                (
                    ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                    ak._v2.contents.bitmaskedarray.BitMaskedArray,
                    ak._v2.contents.unmaskedarray.UnmaskedArray,
                ),
            ):
                rawcontent = self._content.toIndexedOptionArray64()
                inner = rawcontent.index
                result = ak._v2.index.Index64.empty(self.index.length, self._nplike)

            assert (
                result.nplike is self._nplike
                and self._index.nplike is self._nplike
                and inner.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_IndexedArray_simplify",
                    result.dtype.type,
                    self._index.dtype.type,
                    inner.dtype.type,
                ](
                    result.data,
                    self._index.data,
                    self._index.length,
                    inner.data,
                    inner.length,
                )
            )
            if isinstance(self._content, ak._v2.contents.indexedarray.IndexedArray):
                return IndexedArray(
                    result,
                    self._content.content,
                    self._identifier,
                    self._parameters,
                    self._nplike,
                )

            if isinstance(
                self._content,
                (
                    ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                    ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                    ak._v2.contents.bitmaskedarray.BitMaskedArray,
                    ak._v2.contents.unmaskedarray.UnmaskedArray,
                ),
            ):
                return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
                    result,
                    self._content.content,
                    self._identifier,
                    self._parameters,
                    self._nplike,
                )

        else:
            return self

    def num(self, axis, depth=0):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            out = ak._v2.index.Index64.empty(1, self._nplike)
            out[0] = self.length
            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)[
                0
            ]
        else:
            return self.project().num(posaxis, depth)

    def _offsets_and_flattened(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            raise np.AxisError(self, "axis=0 not allowed for flatten")

        else:
            return self.project()._offsets_and_flattened(posaxis, depth)

    def mergeable(self, other, mergebool):
        if not _parameters_equal(self._parameters, other._parameters):
            return False

        if isinstance(
            other,
            (
                ak._v2.contents.emptyarray.EmptyArray,
                ak._v2.contents.unionarray.UnionArray,
            ),
        ):
            return True

        if isinstance(
            other,
            (
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return self._content.mergeable(other.content, mergebool)

        else:
            return self._content.mergeable(other, mergebool)

    def _merging_strategy(self, others):
        if len(others) == 0:
            raise ValueError(
                "to merge this array with 'others', at least one other must be provided"
            )

        head = [self]
        tail = []

        i = 0
        while i < len(others):
            other = others[i]
            if isinstance(other, ak._v2.contents.unionarray.UnionArray):
                break
            else:
                head.append(other)
            i = i + 1

        while i < len(others):
            tail.append(others[i])
            i = i + 1

        if any(
            isinstance(x.nplike, ak._v2._typetracer.TypeTracer) for x in head + tail
        ):
            head = [
                x
                if isinstance(x.nplike, ak._v2._typetracer.TypeTracer)
                else x.typetracer
                for x in head
            ]
            tail = [
                x
                if isinstance(x.nplike, ak._v2._typetracer.TypeTracer)
                else x.typetracer
                for x in tail
            ]

        return (head, tail)

    def _reverse_merge(self, other):
        theirlength = other.length
        mylength = self.length
        index = ak._v2.index.Index64.empty((theirlength + mylength), self._nplike)

        content = other.merge(self._content)

        assert index.nplike is self._nplike
        self._handle_error(
            self._nplike["awkward_IndexedArray_fill_to64_count", index.dtype.type](
                index.data,
                0,
                theirlength,
                0,
            )
        )
        reinterpreted_index = ak._v2.index.Index(
            np.asarray(index.data.view(self[0].dtype))
        )

        assert (
            index.nplike is self._nplike and reinterpreted_index.nplike is self._nplike
        )
        self._handle_error(
            self._nplike[
                "awkward_IndexedArray_fill",
                index.dtype.type,
                reinterpreted_index.dtype.type,
            ](
                index.data,
                theirlength,
                reinterpreted_index.data,
                mylength,
                theirlength,
            )
        )
        parameters = ak._v2._util.merge_parameters(self._parameters, other._parameters)

        return ak._v2.contents.indexedarray.IndexedArray(
            index, content, None, parameters, self._nplike
        )

    def mergemany(self, others):
        if len(others) == 0:
            return self

        head, tail = self._merging_strategy(others)

        total_length = 0
        for array in head:
            total_length += array.length

        contents = []
        contentlength_so_far = 0
        length_so_far = 0
        nextindex = ak._v2.index.Index64.empty(total_length, self._nplike)
        parameters = self._parameters

        for array in head:
            parameters = ak._v2._util.merge_parameters(
                self._parameters, array._parameters, True
            )

            if isinstance(
                array,
                (
                    ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                    ak._v2.contents.bitmaskedarray.BitMaskedArray,
                    ak._v2.contents.unmaskedarray.UnmaskedArray,
                ),
            ):
                array = array.toIndexedOptionArray64()

            if isinstance(array, ak._v2.contents.indexedarray.IndexedArray):
                contents.append(array.content)
                array_index = array.index
                assert (
                    nextindex.nplike is self._nplike
                    and array_index.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_IndexedArray_fill",
                        nextindex.dtype.type,
                        array_index.dtype.type,
                    ](
                        nextindex.data,
                        length_so_far,
                        array_index.data,
                        array.length,
                        contentlength_so_far,
                    )
                )
                contentlength_so_far += array.content.length
                length_so_far += array.length

            elif isinstance(array, ak._v2.contents.emptyarray.EmptyArray):
                pass
            else:
                contents.append(array)
                assert nextindex.nplike is self._nplike
                self._handle_error(
                    self._nplike[
                        "awkward_IndexedArray_fill_count",
                        nextindex.dtype.type,
                    ](
                        nextindex.data,
                        length_so_far,
                        array.length,
                        contentlength_so_far,
                    )
                )
                contentlength_so_far += array.length
                length_so_far += array.length

        tail_contents = contents[1:]
        nextcontent = contents[0].mergemany(tail_contents)
        next = ak._v2.contents.indexedarray.IndexedArray(
            nextindex, nextcontent, None, parameters, self._nplike
        )

        if len(tail) == 0:
            return next

        reversed = tail[0]._reverse_merge(next)
        if len(tail) == 1:
            return reversed
        else:
            return reversed.mergemany(tail[1:])

        raise NotImplementedError(
            "not implemented: " + type(self).__name__ + " ::mergemany"
        )

    def fillna(self, value):
        if value.nplike.known_shape and value.length != 1:
            raise ValueError(f"fillna value length ({value.length}) is not equal to 1")

        return IndexedArray(
            self._index,
            self._content.fillna(value),
            None,
            self._parameters,
            self._nplike,
        )

    def _localindex(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        else:
            return self.project()._localindex(posaxis, depth)

    def _unique_index(self, index, sorted=True):
        next = ak._v2.index.Index64.empty(self.length, self._nplike)
        length = ak._v2.index.Index64.zeros(1, self._nplike)

        if not sorted:
            next = self._index
            offsets = ak._v2.index.Index64.empty(2, self._nplike)
            offsets[0] = 0
            offsets[1] = next.length
            assert next.nplike is self._nplike and offsets.nplike is self._nplike
            self._handle_error(
                self._nplike[
                    "awkward_sort",
                    next.dtype.type,
                    next.dtype.type,
                    offsets.dtype.type,
                ](
                    next.data,
                    next.data,
                    offsets[1],
                    offsets.data,
                    2,
                    offsets[1],
                    True,
                    False,
                )
            )

            assert next.nplike is self._nplike and length.nplike is self._nplike
            self._handle_error(
                self._nplike["awkward_unique", next.dtype.type, length.dtype.type](
                    next.data,
                    self._index.length,
                    length.data,
                )
            )

        else:
            assert (
                self._index.nplike is self._nplike
                and next.nplike is self._nplike
                and length.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_unique_copy",
                    self._index.dtype.type,
                    next.dtype.type,
                    length.dtype.type,
                ](
                    self._index.data,
                    next.data,
                    self._index.length,
                    length.data,
                )
            )

        return next[0 : length[0]]

    def numbers_to_type(self, name):
        return ak._v2.contents.indexedarray.IndexedArray(
            self._index,
            self._content.numbers_to_type(name),
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def _is_unique(self, negaxis, starts, parents, outlength):
        if self._index.length == 0:
            return True

        nextindex = self._unique_index(self._index)

        next = self._content._carry(nextindex, False, NestedIndexError)
        return next._is_unique(negaxis, starts, parents, outlength)

    def _unique(self, negaxis, starts, parents, outlength):
        if self._index.length == 0:
            return self

        branch, depth = self.branch_depth

        index_length = self._index.length
        parents_length = parents.length
        next_length = index_length

        nextcarry = ak._v2.index.Index64.empty(index_length, self._nplike)
        nextparents = ak._v2.index.Index64.empty(index_length, self._nplike)
        outindex = ak._v2.index.Index64.empty(index_length, self._nplike)
        assert (
            nextcarry.nplike is self._nplike
            and nextparents.nplike is self._nplike
            and outindex.nplike is self._nplike
            and self._index.nplike is self._nplike
            and parents.nplike is self._nplike
        )
        self._handle_error(
            self._nplike[
                "awkward_IndexedArray_reduce_next_64",
                nextcarry.dtype.type,
                nextparents.dtype.type,
                outindex.dtype.type,
                self._index.dtype.type,
                parents.dtype.type,
            ](
                nextcarry.data,
                nextparents.data,
                outindex.data,
                self._index.data,
                parents.data,
                index_length,
            )
        )
        next = self._content._carry(nextcarry, False, NestedIndexError)
        unique = next._unique(
            negaxis,
            starts,
            nextparents,
            outlength,
        )

        if branch or (negaxis is not None and negaxis != depth):
            nextoutindex = ak._v2.index.Index64.empty(parents_length, self._nplike)
            assert (
                nextoutindex.nplike is self._nplike
                and starts.nplike is self._nplike
                and parents.nplike is self._nplike
                and nextparents.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_IndexedArray_local_preparenext_64",
                    nextoutindex.dtype.type,
                    starts.dtype.type,
                    parents.dtype.type,
                    nextparents.dtype.type,
                ](
                    nextoutindex.data,
                    starts.data,
                    parents.data,
                    parents_length,
                    nextparents.data,
                    next_length,
                )
            )

            return ak._v2.contents.IndexedOptionArray(
                nextoutindex,
                unique,
                None,
                self._parameters,
                self._nplike,
            ).simplify_optiontype()

        if not branch and negaxis == depth:
            return unique
        else:
            if isinstance(unique, ak._v2.contents.RegularArray):
                unique = unique.toListOffsetArray64(True)

            elif isinstance(unique, ak._v2.contents.ListOffsetArray):
                if starts.nplike.known_data and starts.length > 0 and starts[0] != 0:
                    raise AssertionError(
                        "reduce_next with unbranching depth > negaxis expects a "
                        "ListOffsetArray64 whose offsets start at zero ({})".format(
                            starts[0]
                        )
                    )

                outoffsets = ak._v2.index.Index64.empty(starts.length + 1, self._nplike)
                assert (
                    outoffsets.nplike is self._nplike and starts.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_IndexedArray_reduce_next_fix_offsets_64",
                        outoffsets.dtype.type,
                        starts.dtype.type,
                    ](
                        outoffsets.data,
                        starts.data,
                        starts.length,
                        self._index.length,
                    )
                )

                tmp = ak._v2.contents.IndexedArray(
                    outindex,
                    unique._content,
                    None,
                    None,
                    self._nplike,
                )

                return ak._v2.contents.ListOffsetArray(
                    outoffsets,
                    tmp,
                    None,
                    None,
                    self._nplike,
                )

            elif isinstance(unique, ak._v2.contents.NumpyArray):
                nextoutindex = ak._v2.index.Index64(
                    self._nplike.arange(unique.length, dtype=np.int64)
                )
                out = ak._v2.contents.IndexedOptionArray(
                    nextoutindex,
                    unique,
                    None,
                    self._parameters,
                    self._nplike,
                ).simplify_optiontype()

                return out

        raise NotImplementedError

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self._index.length == 0:
            return ak._v2.contents.NumpyArray(
                self._nplike.empty(0, np.int64), None, None, self._nplike
            )

        next = self._content._carry(self._index, False, NestedIndexError)
        return next._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _sort_next(
        self,
        negaxis,
        starts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        next = self._content._carry(self._index, False, NestedIndexError)
        return next._sort_next(
            negaxis,
            starts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        else:
            return self.project()._combinations(
                n, replacement, recordlookup, parameters, posaxis, depth
            )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        branch, depth = self.branch_depth

        index_length = self._index.length

        nextcarry = ak._v2.index.Index64.empty(index_length, self._nplike)
        nextparents = ak._v2.index.Index64.empty(index_length, self._nplike)
        outindex = ak._v2.index.Index64.empty(index_length, self._nplike)
        assert (
            nextcarry.nplike is self._nplike
            and nextparents.nplike is self._nplike
            and outindex.nplike is self._nplike
            and self._index.nplike is self._nplike
            and parents.nplike is self._nplike
        )
        self._handle_error(
            self._nplike[
                "awkward_IndexedArray_reduce_next_64",
                nextcarry.dtype.type,
                nextparents.dtype.type,
                outindex.dtype.type,
                self._index.dtype.type,
                parents.dtype.type,
            ](
                nextcarry.data,
                nextparents.data,
                outindex.data,
                self._index.data,
                parents.data,
                index_length,
            )
        )
        next = self._content._carry(nextcarry, False, NestedIndexError)
        nextshifts = None
        out = next._reduce_next(
            reducer,
            negaxis,
            starts,
            nextshifts,
            nextparents,
            outlength,
            mask,
            keepdims,
        )

        if not branch and negaxis == depth:
            return out
        else:
            if isinstance(out, ak._v2.contents.RegularArray):
                out = out.toListOffsetArray64(True)

            elif isinstance(out, ak._v2.contents.ListOffsetArray):
                if starts.nplike.known_data and starts.length > 0 and starts[0] != 0:
                    raise AssertionError(
                        "reduce_next with unbranching depth > negaxis expects a "
                        "ListOffsetArray64 whose offsets start at zero ({})".format(
                            starts[0]
                        )
                    )

                outoffsets = ak._v2.index.Index64.empty(starts.length + 1, self._nplike)
                assert (
                    outoffsets.nplike is self._nplike and starts.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_IndexedArray_reduce_next_fix_offsets_64",
                        outoffsets.dtype.type,
                        starts.dtype.type,
                    ](
                        outoffsets.data,
                        starts.data,
                        starts.length,
                        self._index.length,
                    )
                )

                tmp = ak._v2.contents.IndexedArray(
                    outindex,
                    out._content,
                    None,
                    None,
                    self._nplike,
                )

                return ak._v2.contents.ListOffsetArray(
                    outoffsets,
                    tmp,
                    None,
                    None,
                    self._nplike,
                )

            else:
                raise AssertionError(
                    "reduce_next with unbranching depth > negaxis is only "
                    "expected to return RegularArray or ListOffsetArray64; "
                    "instead, it returned " + out.classname
                )

    def _validityerror(self, path):
        error = self._nplike["awkward_IndexedArray_validity", self.index.dtype.type](
            self.index.data, self.index.length, self._content.length, False
        )
        if error.str is not None:
            if error.filename is None:
                filename = ""
            else:
                filename = " (in compiled code: " + error.filename.decode(
                    errors="surrogateescape"
                ).lstrip("\n").lstrip("(")
            message = error.str.decode(errors="surrogateescape")
            return 'at {} ("{}"): {} at i={}{}'.format(
                path, type(self), message, error.id, filename
            )

        elif isinstance(
            self._content,
            (
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return "{0} contains \"{1}\", the operation that made it might have forgotten to call 'simplify_optiontype()'"
        else:
            return self._content.validityerror(path + ".content")

    def _nbytes_part(self):
        result = self.index._nbytes_part() + self.content._nbytes_part()
        if self.identifier is not None:
            result = result + self.identifier._nbytes_part()
        return result

    def _rpad(self, target, axis, depth, clip):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self.rpad_axis0(target, clip)
        elif posaxis == depth + 1:
            return self.project()._rpad(target, posaxis, depth, clip)
        else:
            return ak._v2.contents.indexedarray.IndexedArray(
                self._index,
                self._content._rpad(target, posaxis, depth, clip),
                None,
                self._parameters,
                self._nplike,
            )

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        if (
            not options["categorical_as_dictionary"]
            and self.parameter("__array__") == "categorical"
        ):
            next_parameters = dict(self._parameters)
            del next_parameters["__array__"]
            next = IndexedArray(
                self._index,
                self._content,
                self._identifier,
                next_parameters,
                self._nplike,
            )
            return next._to_arrow(pyarrow, mask_node, validbytes, length, options)

        index = self._index.raw(numpy)

        if self.parameter("__array__") == "categorical":
            dictionary = self._content._to_arrow(
                pyarrow, None, None, self._content.length, options
            )
            out = pyarrow.DictionaryArray.from_arrays(
                index,
                dictionary,
                None if validbytes is None else ~validbytes,
            )
            if options["extensionarray"]:
                return ak._v2._connect.pyarrow.AwkwardArrowArray.from_storage(
                    ak._v2._connect.pyarrow.to_awkwardarrow_type(
                        out.type, options["extensionarray"], mask_node, self
                    ),
                    out,
                )
            else:
                return out

        else:
            if self._content.length == 0:
                # IndexedOptionArray._to_arrow replaces -1 in the index with 0. So behind
                # every masked value is self._content[0], unless self._content.length == 0.
                # In that case, don't call self._content[index]; it's empty anyway.
                next = self._content
            else:
                next = self._content._carry(
                    ak._v2.index.Index(index), False, IndexError
                )

            return next.merge_parameters(self._parameters)._to_arrow(
                pyarrow, mask_node, validbytes, length, options
            )

    def _to_numpy(self, allow_missing):
        return self.project()._to_numpy(allow_missing)

    def _completely_flatten(self, nplike, options):
        return self.project()._completely_flatten(nplike, options)

    def _recursively_apply(
        self, action, depth, depth_context, lateral_context, options
    ):
        if options["return_array"]:

            def continuation():
                return IndexedArray(
                    self._index,
                    self._content._recursively_apply(
                        action,
                        depth,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    ),
                    self._identifier,
                    self._parameters if options["keep_parameters"] else None,
                    self._nplike,
                )

        else:

            def continuation():
                self._content._recursively_apply(
                    action,
                    depth,
                    copy.copy(depth_context),
                    lateral_context,
                    options,
                )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def packed(self):
        return self.project().packed()

    def _to_list(self, behavior):
        out = self._to_list_custom(behavior)
        if out is not None:
            return out

        index = self._index.raw(numpy)
        content = self._content._to_list(behavior)
        out = [None] * index.length
        for i, ind in enumerate(index):
            out[i] = content[ind]
        return out

    def _to_nplike(self, nplike):
        index = self._index._to_nplike(nplike)
        content = self._content._to_nplike(nplike)
        return IndexedArray(
            index,
            content,
            identifier=self.identifier,
            parameters=self.parameters,
            nplike=nplike,
        )

    def _to_json(
        self,
        nan_string,
        infinity_string,
        minus_infinity_string,
        complex_real_string,
        complex_imag_string,
    ):
        out = self._to_json_custom()
        if out is not None:
            return out

        index = self._index.raw(numpy)
        content = self._content._to_json(
            nan_string,
            infinity_string,
            minus_infinity_string,
            complex_real_string,
            complex_imag_string,
        )
        out = [None] * index.length
        for i, ind in enumerate(index):
            out[i] = content[ind]
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
import json
from numbers import Number

np = ak.nplike.NumpyMetadata.instance()


def to_json_file(
    array,
    destination=None,
    pretty=False,
    maxdecimals=None,
    nan_string=None,
    infinity_string=None,
    minus_infinity_string=None,
    complex_record_fields=None,
    buffersize=65536,
):
    """
    Args:
        array: Data to convert to JSON.
        destination (str): a file name to write to (overwrite) that
            file (returning None).
        pretty (bool): If True, indent the output for human readability; if
            False, output compact JSON without spaces.
        maxdecimals (None or int): If an int, limit the number of
            floating-point decimals to this number; if None, write all digits.
        nan_string (None or str): If not None, floating-point NaN values will be
            replaced with this string instead of a JSON number.
        infinity_string (None or str): If not None, floating-point positive infinity
            values will be replaced with this string instead of a JSON number.
        minus_infinity_string (None or str): If not None, floating-point negative
            infinity values will be replaced with this string instead of a JSON
            number.
        complex_record_fields (None or (str, str)): If not None, defines a pair of
            field names to interpret records as complex numbers.
        buffersize (int): Size (in bytes) of the buffer used by the JSON
            parser.

    Converts `array` (many types supported, including all Awkward Arrays and
    Records) into a JSON file.

    Awkward Array types have the following JSON translations.

       * #ak.types.PrimitiveType: converted into JSON booleans and numbers.
       * #ak.types.OptionType: missing values are converted into None.
       * #ak.types.ListType: converted into JSON lists.
       * #ak.types.RegularType: also converted into JSON lists. JSON (and
         Python) forms lose information about the regularity of list lengths.
       * #ak.types.ListType with parameter `"__array__"` equal to
         `"__bytestring__"` or `"__string__"`: converted into JSON strings.
       * #ak.types.RecordArray without field names: converted into JSON
         objects with numbers as strings for keys.
       * #ak.types.RecordArray with field names: converted into JSON objects.
       * #ak.types.UnionArray: JSON data are naturally heterogeneous.

    See also #ak.from_json and #ak.Array.tojson.
    """
    if array is None or isinstance(array, (bool, str, bytes, Number)):
        return json.dump(array)

    elif isinstance(array, bytes):
        return json.dump(array.decode("utf-8", "surrogateescape"))

    elif isinstance(array, np.ndarray):
        out = ak._v2.contents.NumpyArray(array)

    elif isinstance(array, ak._v2.highlevel.Array):
        out = array.layout

    elif isinstance(array, ak._v2.highlevel.Record):
        out = array.layout

    elif isinstance(array, ak._v2.highlevel.ArrayBuilder):
        out = array.snapshot().layout

    elif isinstance(array, ak._v2.record.Record):
        out = array

    elif isinstance(array, ak.layout.ArrayBuilder):
        formstr, length, buffers = array.to_buffers()
        form = ak._v2.forms.from_json(formstr)

        out = ak._v2.operations.convert.from_buffers(
            form, length, buffers, highlevel=False
        )
        # FIXME: the code is a copy from snapshot,
        # because this call returns v1:
        # out = array.snapshot()

    elif isinstance(array, ak._v2.contents.Content):
        out = array

    else:
        raise TypeError(f"unrecognized array type: {repr(array)}")

    if complex_record_fields is None:
        complex_real_string = None
        complex_imag_string = None
    elif (
        isinstance(complex_record_fields, tuple)
        and len(complex_record_fields) == 2
        and isinstance(complex_record_fields[0], str)
        and isinstance(complex_record_fields[1], str)
    ):
        complex_real_string, complex_imag_string = complex_record_fields

    with open(destination, "w", encoding="utf-8") as f:
        for chunk in json.dumps(
            out.tojson(
                nan_string,
                infinity_string,
                minus_infinity_string,
                complex_real_string,
                complex_imag_string,
            ),
            separators=(",", ":"),
        ):
            f.write(chunk)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_array_builder():
    builder = ak.ArrayBuilder()

    builder.real(1.1)
    builder.real(2.2)
    builder.real(3.3)

    assert ak.to_list(builder.snapshot()) == [1.1, 2.2, 3.3]


def test_array_bool_builder():
    builder = ak.ArrayBuilder()

    builder.boolean(True)
    builder.boolean(False)
    builder.boolean(True)

    assert ak.to_list(builder.snapshot()) == [True, False, True]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("ravel")
def ravel(array, highlevel=True, behavior=None):

    """
    Args:
        array: Data containing nested lists to flatten
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array with all level of nesting removed by erasing the
    boundaries between consecutive lists.

    This is the equivalent of NumPy's `np.ravel` for Awkward Arrays.

    Consider the following doubly nested `array`.

        ak.Array([[
                   [1.1, 2.2, 3.3],
                   [],
                   [4.4, 5.5],
                   [6.6]],
                  [],
                  [
                   [7.7],
                   [8.8, 9.9]
                  ]])

    Ravelling the array produces a flat array

        >>> print(ak.ravel(array))
        [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    Missing values are eliminated by flattening: there is no distinction
    between an empty list and a value of None at the level of flattening.
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    nplike = ak.nplike.of(layout)

    out = layout.completely_flatten(function_name="ak.ravel")
    assert isinstance(out, tuple) and all(isinstance(x, np.ndarray) for x in out)

    if any(isinstance(x, nplike.ma.MaskedArray) for x in out):
        out = ak._v2.contents.NumpyArray(nplike.ma.concatenate(out))
    else:
        out = ak._v2.contents.NumpyArray(nplike.concatenate(out))

    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("std")
def std(
    x,
    weight=None,
    ddof=0,
    axis=None,
    keepdims=False,
    mask_identity=True,
    flatten_records=False,
):
    """
    Args:
        x: The data on which to compute the standard deviation (anything #ak.to_layout recognizes).
        weight: Data that can be broadcasted to `x` to give each value a
            weight. Weighting values equally is the same as no weights;
            weighting some values higher increases the significance of those
            values. Weights can be zero or negative.
        ddof (int): "delta degrees of freedom": the divisor used in the
            calculation is `sum(weights) - ddof`. Use this for "reduced
            standard deviation."
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this function decreases the number of
            dimensions by 1; if True, the output values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, the application of this function on
            empty lists results in None (an option type); otherwise, the
            calculation is followed through with the reducers' identities,
            usually resulting in floating-point `nan`.
        flatten_records (bool): If True, axis=None combines fields from different
            records; otherwise, records raise an error.

    Computes the standard deviation in each group of elements from `x`
    (many types supported, including all Awkward Arrays and Records). The
    grouping is performed the same way as for reducers, though this operation
    is not a reducer and has no identity. It is the same as NumPy's
    [std](https://docs.scipy.org/doc/numpy/reference/generated/numpy.std.html)
    if all lists at a given dimension have the same length and no None values,
    but it generalizes to cases where they do not.

    Passing all arguments to the reducers, the standard deviation is
    calculated as

        np.sqrt(ak.var(x, weight))

    See #ak.sum for a complete description of handling nested lists and
    missing values (None) in reducers, and #ak.mean for an example with another
    non-reducer.
    """
    x = ak._v2.highlevel.Array(
        ak._v2.operations.convert.to_layout(x, allow_record=False, allow_other=False)
    )
    if weight is not None:
        weight = ak._v2.highlevel.Array(
            ak._v2.operations.convert.to_layout(
                weight, allow_record=False, allow_other=False
            )
        )

    with np.errstate(invalid="ignore"):
        return ak.nplike.of(x, weight).sqrt(
            ak._v2.operations.reducers.var(
                x,
                weight=weight,
                ddof=ddof,
                axis=axis,
                keepdims=keepdims,
                mask_identity=mask_identity,
            )
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def test():
    empty1 = ak._v2.highlevel.Array(ak._v2.contents.EmptyArray(), check_valid=True)
    empty2 = ak._v2.highlevel.Array(
        ak._v2.contents.ListOffsetArray(
            ak._v2.index.Index64(np.array([0, 0, 0, 0], dtype=np.int64)),
            ak._v2.contents.EmptyArray(),
        ),
        check_valid=False,
    )
    array = ak._v2.highlevel.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True)

    assert ak._v2.operations.convert.to_numpy(empty1).dtype.type is np.float64

    assert to_list(array[empty1]) == []
    assert (
        to_list(
            array[
                empty1,
            ]
        )
        == []
    )
    assert to_list(array[empty2]) == [[], [], []]
    assert to_list(
        array[
            empty2,
        ]
    ) == [[], [], []]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
from awkward._v2.types.type import Type
from awkward._v2.types.regulartype import RegularType
from awkward._v2.types.listtype import ListType
from awkward._v2.types.uniontype import UnionType
from awkward._v2.forms.form import _parameters_equal


class OptionType(Type):
    def __init__(self, content, parameters=None, typestr=None):
        if not isinstance(content, Type):
            raise TypeError(
                "{} 'content' must be a Type subclass, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                "{} 'parameters' must be of type dict or None, not {}".format(
                    type(self).__name__, repr(parameters)
                )
            )
        if typestr is not None and not ak._util.isstr(typestr):
            raise TypeError(
                "{} 'typestr' must be of type string or None, not {}".format(
                    type(self).__name__, repr(typestr)
                )
            )
        self._content = content
        self._parameters = parameters
        self._typestr = typestr

    @property
    def content(self):
        return self._content

    def __str__(self):
        if self._typestr is not None:
            out = self._typestr

        else:
            params = self._str_parameters()
            if params is None:
                if not isinstance(self._content, (RegularType, ListType)):
                    out = f"?{str(self._content)}"
                else:
                    out = f"option[{str(self._content)}]"
            else:
                out = f"option[{str(self._content)}, {params}]"

        return self._str_categorical_begin() + out + self._str_categorical_end()

    def __repr__(self):
        args = [repr(self._content)] + self._repr_args()
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def __eq__(self, other):
        if isinstance(other, OptionType):
            return (
                self._typestr == other._typestr
                and _parameters_equal(
                    self._parameters, other._parameters, only_array_record=True
                )
                and self._content == other._content
            )
        else:
            return False

    def simplify_option_union(self):
        if isinstance(self._content, UnionType):
            contents = []
            for content in self._content.contents:
                if isinstance(content, OptionType):
                    contents.append(
                        OptionType(
                            content.content,
                            ak._v2._util.merge_parameters(
                                self._parameters, content._parameters
                            ),
                            content._typestr
                            if self._typestr is None
                            else self._typestr,
                        )
                    )

                else:
                    contents.append(
                        OptionType(content, self._parameters, self._typestr)
                    )

            return UnionType(contents, self._content.parameters, self._content.typestr)

        else:
            return self



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_index32():
    py_array = ak._v2.index.Index.zeros(10, np, np.int32)

    assert len(py_array) == 10
    assert "i32" == py_array.form


def test_index64():
    py_array = ak._v2.index.Index.zeros(10, np, np.int64)

    assert len(py_array) == 10
    assert "i64" == py_array.form


def test_identifier32():
    py_array = ak._v2.identifier.Identifier.zeros(
        123, {1: "one", 2: "two"}, 5, 10, np, np.int32
    )

    assert len(py_array) == 5
    assert py_array._data.dtype == np.dtype(np.int32)


def test_identifier64():
    py_array = ak._v2.identifier.Identifier.zeros(
        123, {1: "one", 2: "two"}, 5, 10, np, np.int64
    )

    assert len(py_array) == 5
    assert py_array._data.dtype == np.dtype(np.int64)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    x = ak._v2.highlevel.Array([1, 2, 3, None, 4])
    assert ak._v2.operations.reducers.argmax(x) == 4



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from awkward._v2.operations.io.ak_from_json_file import from_json_file  # noqa: F401
from awkward._v2.operations.io.ak_to_json_file import to_json_file  # noqa: F401
from awkward._v2.operations.io.ak_from_parquet import from_parquet  # noqa: F401
from awkward._v2.operations.io.ak_to_parquet import to_parquet  # noqa: F401



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_example():
    x = ak._v2.operations.convert.from_iter([[1, 2, 3, None], [], [4, 5]])
    y = ak._v2.operations.convert.from_iter([100, 200, 300])

    ttx = ak._v2.highlevel.Array(x.layout.typetracer)
    tty = ak._v2.highlevel.Array(y.layout.typetracer)

    assert (x + y).layout.form == (ttx + tty).layout.form
    assert (x + np.sin(y)).layout.form == (ttx + np.sin(tty)).layout.form

    x = ak._v2.highlevel.Array(
        ak._v2.contents.ListArray(x.layout.starts, x.layout.stops, x.layout.content)
    )
    ttx = ak._v2.highlevel.Array(x.layout.typetracer)

    assert (x + x).layout.form == (ttx + ttx).layout.form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def test():
    assert to_list(
        ak._v2.operations.reducers.prod(
            ak._v2.highlevel.Array([[[2, 3, 5]], [[7], [11]], [[]]]), axis=-1
        )
    ) == [
        [30],
        [7, 11],
        [1],
    ]

    assert to_list(
        ak._v2.operations.reducers.prod(
            ak._v2.highlevel.Array([[[2, 3, 5]], [[7], [11]], []]), axis=-1
        )
    ) == [
        [30],
        [7, 11],
        [],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def unflatten(array, counts, axis=0, highlevel=True, behavior=None):
    raise NotImplementedError


#     """
#     Args:
#         array: Data to create an array with an additional level from.
#         counts (int or array): Number of elements the new level should have.
#             If an integer, the new level will be regularly sized; otherwise,
#             it will consist of variable-length lists with the given lengths.
#         axis (int): The dimension at which this operation is applied. The
#             outermost dimension is `0`, followed by `1`, etc., and negative
#             values count backward from the innermost: `-1` is the innermost
#             dimension, `-2` is the next level up, etc.
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.
#         behavior (None or dict): Custom #ak.behavior for the output array, if
#             high-level.

#     Returns an array with an additional level of nesting. This is roughly the
#     inverse of #ak.flatten, where `counts` were obtained by #ak.num (both with
#     `axis=1`).

#     For example,

#         >>> original = ak.Array([[0, 1, 2], [], [3, 4], [5], [6, 7, 8, 9]])
#         >>> counts = ak.num(original)
#         >>> array = ak.flatten(original)
#         >>> counts
#         <Array [3, 0, 2, 1, 4] type='5 * int64'>
#         >>> array
#         <Array [0, 1, 2, 3, 4, 5, 6, 7, 8, 9] type='10 * int64'>
#         >>> ak.unflatten(array, counts)
#         <Array [[0, 1, 2], [], ... [5], [6, 7, 8, 9]] type='5 * var * int64'>

#     An inner dimension can be unflattened by setting the `axis` parameter, but
#     operations like this constrain the `counts` more tightly.

#     For example, we can subdivide an already divided list:

#         >>> original = ak.Array([[1, 2, 3, 4], [], [5, 6, 7], [8, 9]])
#         >>> print(ak.unflatten(original, [2, 2, 1, 2, 1, 1], axis=1))
#         [[[1, 2], [3, 4]], [], [[5], [6, 7]], [[8], [9]]]

#     But the counts have to add up to the lengths of those lists. We can't mix
#     values from the first `[1, 2, 3, 4]` with values from the next `[5, 6, 7]`.

#         >>> print(ak.unflatten(original, [2, 1, 2, 2, 1, 1], axis=1))
#         Traceback (most recent call last):
#         ...
#         ValueError: structure imposed by 'counts' does not fit in the array at axis=1

#     Also note that new lists created by this function cannot cross partitions
#     (which is only possible at `axis=0`, anyway).

#     See also #ak.num and #ak.flatten.
#     """
#     nplike = ak.nplike.of(array)

#     layout = ak._v2.operations.convert.to_layout(
#         array, allow_record=False, allow_other=False
#     )

#     if isinstance(counts, (numbers.Integral, np.integer)):
#         current_offsets = None
#     else:
#         counts = ak._v2.operations.convert.to_layout(
#             counts, allow_record=False, allow_other=False
#         )
#         ptr_lib = ak._v2.operations.convert.kernels(array)
#         counts = ak._v2.operations.convert.to_kernels(counts, ptr_lib, highlevel=False)
#         if ptr_lib == "cpu":
#             counts = ak._v2.operations.convert.to_numpy(counts, allow_missing=True)
#             mask = ak.nplike.numpy.ma.getmask(counts)
#             counts = ak.nplike.numpy.ma.filled(counts, 0)
#         elif ptr_lib == "cuda":
#             counts = ak._v2.operations.convert.to_cupy(counts)
#             mask = False
#         else:
#             raise AssertionError(
#                 "unrecognized kernels lib"
#             )
#         if counts.ndim != 1:
#             raise ValueError(
#                 "counts must be one-dimensional"
#             )
#         if not issubclass(counts.dtype.type, np.integer):
#             raise ValueError(
#                 "counts must be integers"
#             )
#         current_offsets = [nplike.empty(len(counts) + 1, np.int64)]
#         current_offsets[0][0] = 0
#         nplike.cumsum(counts, out=current_offsets[0][1:])

#     def doit(layout):
#         if isinstance(counts, (numbers.Integral, np.integer)):
#             if counts < 0 or counts > len(layout):
#                 raise ValueError(
#                     "too large counts for array or negative counts"
#
#                 )
#             out = ak._v2.contents.RegularArray(layout, counts)

#         else:
#             position = (
#                 nplike.searchsorted(
#                     current_offsets[0], nplike.array([len(layout)]), side="right"
#                 )[0]
#                 - 1
#             )
#             if position >= len(current_offsets[0]) or current_offsets[0][
#                 position
#             ] != len(layout):
#                 raise ValueError(
#                     "structure imposed by 'counts' does not fit in the array or partition "
#                     "at axis={0}".format(axis)
#                 )

#             offsets = current_offsets[0][: position + 1]
#             current_offsets[0] = current_offsets[0][position:] - len(layout)

#             out = ak._v2.contents.ListOffsetArray64(ak._v2.contents.Index64(offsets), layout)
#             if not isinstance(mask, (bool, np.bool_)):
#                 index = ak._v2.index.Index8(nplike.asarray(mask).astype(np.int8))
#                 out = ak._v2.contents.ByteMaskedArray(index, out, valid_when=False)

#         return out

#     if axis == 0 or layout.axis_wrap_if_negative(axis) == 0:
#         if isinstance(layout, ak.partition.PartitionedArray):   # NO PARTITIONED ARRAY
#             outparts = []
#             for part in layout.partitions:
#                 outparts.append(doit(part))
#             out = ak.partition.IrregularlyPartitionedArray(outparts)   # NO PARTITIONED ARRAY
#         else:
#             out = doit(layout)

#     else:

#         def transform(layout, depth, posaxis):
#             # Pack the current layout. This ensures that the `counts` array,
#             # which is computed with these layouts applied, aligns with the
#             # internal layout to be unflattened (#910)
#             layout = _pack_layout(layout)

#             posaxis = layout.axis_wrap_if_negative(posaxis)
#             if posaxis == depth and isinstance(layout, ak._v2._util.listtypes):
#                 # We are one *above* the level where we want to apply this.
#                 listoffsetarray = layout.toListOffsetArray64(True)
#                 outeroffsets = nplike.asarray(listoffsetarray.offsets)

#                 content = doit(listoffsetarray.content[: outeroffsets[-1]])
#                 if isinstance(content, ak._v2.contents.ByteMaskedArray):
#                     inneroffsets = nplike.asarray(content.content.offsets)
#                 elif isinstance(content, ak._v2.contents.RegularArray):
#                     inneroffsets = nplike.asarray(
#                         content.toListOffsetArray64(True).offsets
#                     )
#                 else:
#                     inneroffsets = nplike.asarray(content.offsets)

#                 positions = (
#                     nplike.searchsorted(inneroffsets, outeroffsets, side="right") - 1
#                 )
#                 if not nplike.array_equal(inneroffsets[positions], outeroffsets):
#                     raise ValueError(
#                         "structure imposed by 'counts' does not fit in the array or partition "
#                         "at axis={0}".format(axis)
#                     )
#                 positions[0] = 0

#                 return ak._v2.contents.ListOffsetArray64(
#                     ak._v2.index.Index64(positions), content
#                 )

#             else:
#                 return ak._v2._util.transform_child_layouts(
#                     transform, layout, depth, posaxis
#                 )

#         if isinstance(layout, ak.partition.PartitionedArray):   # NO PARTITIONED ARRAY
#             outparts = []
#             for part in layout.partitions:
#                 outparts.append(transform(part, depth=1, posaxis=axis))
#             out = ak.partition.IrregularlyPartitionedArray(outparts)   # NO PARTITIONED ARRAY
#         else:
#             out = transform(layout, depth=1, posaxis=axis)

#     if current_offsets is not None and not (
#         len(current_offsets[0]) == 1 and current_offsets[0][0] == 0
#     ):
#         raise ValueError(
#             "structure imposed by 'counts' does not fit in the array or partition "
#             "at axis={0}".format(axis)
#         )

#     return ak._v2._util.maybe_wrap_like(out, array, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from awkward._v2.operations.convert.ak_from_numpy import from_numpy  # noqa: F401
from awkward._v2.operations.convert.ak_to_numpy import to_numpy  # noqa: F401
from awkward._v2.operations.convert.ak_from_cupy import from_cupy  # noqa: F401
from awkward._v2.operations.convert.ak_to_cupy import to_cupy  # noqa: F401
from awkward._v2.operations.convert.ak_from_jax import from_jax  # noqa: F401
from awkward._v2.operations.convert.ak_to_jax import to_jax  # noqa: F401
from awkward._v2.operations.convert.ak_from_iter import from_iter  # noqa: F401
from awkward._v2.operations.convert.ak_to_list import to_list  # noqa: F401
from awkward._v2.operations.convert.ak_from_json import from_json  # noqa: F401
from awkward._v2.operations.convert.ak_from_json_schema import (  # noqa: F401
    from_json_schema,
)
from awkward._v2.operations.convert.ak_to_json import to_json  # noqa: F401
from awkward._v2.operations.convert.ak_to_layout import to_layout  # noqa: F401
from awkward._v2.operations.convert.ak_to_arrow import to_arrow  # noqa: F401
from awkward._v2.operations.convert.ak_to_arrow_table import (  # noqa: F401
    to_arrow_table,
)
from awkward._v2.operations.convert.ak_from_arrow import from_arrow  # noqa: F401
from awkward._v2.operations.convert.ak_from_arrow_schema import (  # noqa: F401
    from_arrow_schema,
)
from awkward._v2.operations.convert.ak_to_parquet import to_parquet  # noqa: F401
from awkward._v2.operations.convert.ak_from_parquet import from_parquet  # noqa: F401
from awkward._v2.operations.convert.ak_to_buffers import to_buffers  # noqa: F401
from awkward._v2.operations.convert.ak_from_buffers import from_buffers  # noqa: F401
from awkward._v2.operations.convert.ak_to_pandas import to_pandas  # noqa: F401



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def mask(array, mask, valid_when=True, highlevel=True, behavior=None):

    """
    Args:
        array: Data to mask, rather than filter.
        mask (array of booleans): The mask that overlays elements in the
            `array` with None. Must have the same length as `array`.
        valid_when (bool): If True, True values in `mask` are considered
            valid (passed from `array` to the output); if False, False
            values in `mask` are considered valid.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array for which

        output[i] = array[i] if mask[i] == valid_when else None

    Unlike filtering data with #ak.Array.__getitem__, this `output` has the
    same length as the original `array` and can therefore be used in
    calculations with it, such as
    [universal functions](https://docs.scipy.org/doc/numpy/reference/ufuncs.html).

    For example, with an `array` like

        ak.Array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

    with a boolean selection of `good` elements like

        >>> good = (array % 2 == 1)
        >>> good
        <Array [False, True, False, ... False, True] type='10 * bool'>

    could be used to filter the original `array` (or another with the same
    length).

        >>> array[good]
        <Array [1, 3, 5, 7, 9] type='5 * int64'>

    However, this eliminates information about which elements were dropped and
    where they were. If we instead use #ak.mask,

        >>> ak.mask(array, good)
        <Array [None, 1, None, 3, ... None, 7, None, 9] type='10 * ?int64'>

    this information and the length of the array is preserved, and it can be
    used in further calculations with the original `array` (or another with
    the same length).

        >>> ak.mask(array, good) + array
        <Array [None, 2, None, 6, ... 14, None, 18] type='10 * ?int64'>

    In particular, successive filters can be applied to the same array.

    Even if the `array` and/or the `mask` is nested,

        >>> array = ak.Array([[[0, 1, 2], [], [3, 4], [5]], [[6, 7, 8], [9]]])
        >>> good = (array % 2 == 1)
        >>> good
        <Array [[[False, True, False], ... [True]]] type='2 * var * var * bool'>

    it can still be used with #ak.mask because the `array` and `mask`
    parameters are broadcasted.

        >>> ak.mask(array, good)
        <Array [[[None, 1, None], ... None], [9]]] type='2 * var * var * ?int64'>

    See #ak.broadcast_arrays for details about broadcasting and the generalized
    set of broadcasting rules.

    Another syntax for

        ak.mask(array, array_of_booleans)

    is

        array.mask[array_of_booleans]

    (which is 5 characters away from simply filtering the `array`).
    """

    def action(inputs, **kwargs):
        layoutarray, layoutmask = inputs
        if isinstance(layoutmask, ak._v2.contents.NumpyArray):
            m = ak.nplike.of(layoutmask).asarray(layoutmask)
            if not issubclass(m.dtype.type, (bool, np.bool_)):
                raise ValueError(
                    "mask must have boolean type, not " "{}".format(repr(m.dtype))
                )
            bytemask = ak._v2.index.Index8(m.view(np.int8))
            return (
                ak._v2.contents.ByteMaskedArray(
                    bytemask, layoutarray, valid_when=valid_when
                ).simplify_optiontype(),
            )
        else:
            return None

    layoutarray = ak._v2.operations.convert.to_layout(
        array, allow_record=True, allow_other=False
    )
    layoutmask = ak._v2.operations.convert.to_layout(
        mask, allow_record=True, allow_other=False
    )

    behavior = ak._v2._util.behavior_of(array, mask, behavior=behavior)
    out = ak._v2._broadcasting.broadcast_and_apply(
        [layoutarray, layoutmask],
        action,
        behavior,
        numpy_to_regular=True,
        right_broadcast=False,
    )
    assert isinstance(out, tuple) and len(out) == 1
    return ak._v2._util.wrap(out[0], behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def test_unknown():
    i = ak._v2.index.Index64(np.array([0, 0, 0, 0], dtype=np.int64))
    e = ak._v2.contents.EmptyArray()
    a = ak._v2.contents.ListOffsetArray(i, e)
    assert to_list(a) == [[], [], []]
    assert str(ak._v2.operations.describe.type(a)) == "var * unknown"
    assert ak._v2.operations.describe.type(a) == ak._v2.types.ListType(
        ak._v2.types.UnknownType()
    )
    assert not ak._v2.operations.describe.type(a) == ak._v2.types.NumpyType("float64")

    i = ak._v2.index.Index64(np.array([0, 0, 0, 0, 0, 0], dtype=np.int64))
    ii = ak._v2.index.Index64(np.array([0, 0, 2, 5], dtype=np.int64))
    a = ak._v2.contents.ListOffsetArray(i, e)
    a = ak._v2.contents.ListOffsetArray(ii, a)

    assert to_list(a) == [[], [[], []], [[], [], []]]
    assert str(ak._v2.operations.describe.type(a)) == "var * var * unknown"
    assert ak._v2.operations.describe.type(a) == ak._v2.types.ListType(
        ak._v2.types.ListType(ak._v2.types.UnknownType())
    )


@pytest.mark.skip(
    reason="FIXME:  ArrayType(ListType(UnknownType()), 3) == ListType(UnknownType())"
)
def test_unknown_arraybuilder():
    a = ak._v2.highlevel.ArrayBuilder()
    a.begin_list()
    a.end_list()
    a.begin_list()
    a.end_list()
    a.begin_list()
    a.end_list()
    assert to_list(a) == [[], [], []]
    assert str(ak._v2.operations.describe.type(a)) == "3 * var * unknown"
    assert ak._v2.operations.describe.type(a) == ak._v2.types.ListType(
        ak._v2.types.UnknownType()
    )
    assert not ak._v2.operations.describe.type(a) == ak._v2.types.NumpyType("float64")

    a = a.snapshot()
    assert to_list(a) == [[], [], []]
    assert str(ak._v2.operations.describe.type(a)) == "var * unknown"
    assert ak._v2.operations.describe.type(a) == ak._v2.types.ListType(
        ak._v2.types.UnknownType()
    )
    assert not ak._v2.operations.describe.type(a) == ak._v2.types.NumpyType("float64")


def test_getitem():
    a = ak._v2.operations.convert.from_iter(
        [[], [[], []], [[], [], []]], highlevel=False
    )

    assert to_list(a[2]) == [[], [], []]
    assert a.typetracer[2].form == a[2].form

    assert to_list(a[2, 1]) == []
    assert a.typetracer[2, 1].form == a[2, 1].form
    with pytest.raises(ValueError):
        a[2, 1, 0]
    assert to_list(a[2, 1][()]) == []
    assert a.typetracer[2, 1][()].form == a[2, 1][()].form
    with pytest.raises(IndexError):
        a[2, 1][0]
    assert to_list(a[2, 1][100:200]) == []
    assert a.typetracer[2, 1][100:200].form == a[2, 1][100:200].form
    assert to_list(a[2, 1, 100:200]) == []
    assert a.typetracer[2, 1, 100:200].form == a[2, 1, 100:200].form
    assert to_list(a[2, 1][np.array([], dtype=int)]) == []
    assert (
        a.typetracer[2, 1][np.array([], dtype=int)].form
        == a[2, 1][np.array([], dtype=int)].form
    )
    assert to_list(a[2, 1, np.array([], dtype=int)]) == []
    assert (
        a.typetracer[2, 1, np.array([], dtype=int)].form
        == a[2, 1, np.array([], dtype=int)].form
    )
    with pytest.raises(ValueError):
        a[2, 1, np.array([0], dtype=int)]
    with pytest.raises(IndexError):
        a[2, 1][100:200, 0]
    with pytest.raises(IndexError):
        a[2, 1][100:200, 200:300]

    assert to_list(a[2, 1][100:200, np.array([], dtype=int)]) == []
    assert (
        a.typetracer[2, 1][100:200, np.array([], dtype=int)].form
        == a[2, 1][100:200, np.array([], dtype=int)].form
    )

    assert to_list(a[1:, 1:]) == [[[]], [[], []]]
    assert a.typetracer[1:, 1:].form == a[1:, 1:].form
    with pytest.raises(ValueError):
        a[1:, 1:, 0]


def test_unknown2():
    a = ak._v2.operations.convert.from_json("[[], [], []]", highlevel=False)
    assert a.tolist() == [[], [], []]
    assert str(a.form.type) == "var * unknown"
    assert a.form.type == ak._v2.types.ListType(ak._v2.types.UnknownType())
    assert not a.form.type == ak._v2.types.NumpyType("float64")

    a = ak._v2.operations.convert.from_json(
        "[[], [[], []], [[], [], []]]", highlevel=False
    )
    assert a.tolist() == [[], [[], []], [[], [], []]]
    assert str(a.form.type) == "var * var * unknown"
    assert a.form.type == ak._v2.types.ListType(
        ak._v2.types.ListType(ak._v2.types.UnknownType())
    )

    a = ak._v2.highlevel.ArrayBuilder()
    a.begin_list()
    a.end_list()
    a.begin_list()
    a.end_list()
    a.begin_list()
    a.end_list()
    assert a.tolist() == [[], [], []]
    assert str(a.type) == "3 * var * unknown"
    assert a.type == ak._v2.types.ArrayType(
        ak._v2.types.ListType(ak._v2.types.UnknownType()), 3
    )
    assert not a.type == ak._v2.types.NumpyType("float64")

    a = a.snapshot()
    assert a.tolist() == [[], [], []]
    assert str(a.type) == "3 * var * unknown"
    assert a.type == ak._v2.types.ArrayType(
        ak._v2.types.ListType(ak._v2.types.UnknownType()), 3
    )
    assert not a.type == ak._v2.types.NumpyType("float64")


def test_from_json_getitem():
    a = ak._v2.operations.convert.from_json("[]")
    a = ak._v2.operations.convert.from_json("[[], [[], []], [[], [], []]]")
    assert a[2].tolist() == [[], [], []]

    assert a[2, 1].tolist() == []
    with pytest.raises(ValueError) as excinfo:
        a[2, 1, 0]
    assert "index out of range while attempting to get index 0" in str(excinfo.value)
    assert a[2, 1][()].tolist() == []
    with pytest.raises(IndexError) as excinfo:
        a[2, 1][0]
    assert (
        "cannot slice\n\n    <Array [] type='0 * unknown'>\n\nwith\n\n    0\n\nat inner EmptyArray of length 0, using sub-slice 0.\n\nEmptyArray error: array is empty."
        in str(excinfo.value)
    )
    assert a[2, 1][100:200].tolist() == []
    assert a[2, 1, 100:200].tolist() == []
    assert a[2, 1][np.array([], dtype=int)].tolist() == []
    assert a[2, 1, np.array([], dtype=int)].tolist() == []
    with pytest.raises(ValueError) as excinfo:
        a[2, 1, np.array([0], dtype=int)]
    assert "index out of range while attempting to get index 0" in str(excinfo.value)
    with pytest.raises(IndexError) as excinfo:
        a[2, 1][100:200, 0]
    assert (
        "cannot slice\n\n    <Array [] type='0 * unknown'>\n\nwith\n\n    (100:200, 0)\n\nat inner EmptyArray of length 0, using sub-slice 0.\n\nEmptyArray error: array is empty."
        in str(excinfo.value)
    )
    with pytest.raises(IndexError) as excinfo:
        a[2, 1][100:200, 200:300]
    assert (
        "cannot slice\n\n    <Array [] type='0 * unknown'>\n\nwith\n\n    (100:200, 200:300)\n\nat inner EmptyArray of length 0, using sub-slice 200:300.\n\nEmptyArray error: array is empty."
        in str(excinfo.value)
    )

    # FIXME: Failed: DID NOT RAISE <class 'IndexError'>
    # with pytest.raises(ValueError) as excinfo:
    #     a[2, 1][100:200, np.array([], dtype=int)]
    # assert ", too many dimensions in slice" in str(excinfo.value)

    assert a[1:, 1:].tolist() == [[[]], [[], []]]
    with pytest.raises(ValueError) as excinfo:
        a[1:, 1:, 0]
    assert "index out of range while attempting to get index 0" in str(excinfo.value)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from awkward._v2.operations.structure.ak_zeros_like import _ZEROS
import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("full_like")
def full_like(array, fill_value, highlevel=True, behavior=None, dtype=None):

    """
    Args:
        array: Array to use as a model for a replacement that contains only
            `fill_value`.
        fill_value: Value to fill new new array with.
        highlevel (bool, default is True): If True, return an #ak.Array;
            otherwise, return a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        dtype (None or NumPy dtype)): Overrides the data type of the result.

    This is the equivalent of NumPy's `np.full_like` for Awkward Arrays.

    Although it's possible to produce an array of `fill_value` with the structure
    of an `array` using #ak.broadcast_arrays:

        >>> array = ak.Array([[1, 2, 3], [], [4, 5]])
        >>> ak.broadcast_arrays(array, 1)
        [<Array [[1, 2, 3], [], [4, 5]] type='3 * var * int64'>,
         <Array [[1, 1, 1], [], [1, 1]] type='3 * var * int64'>]
        >>> ak.broadcast_arrays(array, 1.0)
        [<Array [[1, 2, 3], [], [4, 5]] type='3 * var * int64'>,
         <Array [[1, 1, 1], [], [1, 1]] type='3 * var * float64'>]

    Such a technique takes its type from the scalar (`1` or `1.0`), rather than
    the array. This function gets all types from the array, which might not be
    the same in all parts of the structure.

    Here is an extreme example:

        >>> array = ak.Array([
        ... [{"x": 0.0, "y": []},
        ...  {"x": 1.1, "y": [1]},
        ...  {"x": 2.2, "y": [1, 2]}],
        ... [],
        ... [{"x": 3.3, "y": [1, 2, None, 3]},
        ...  False,
        ...  False,
        ...  True,
        ...  {"x": 4.4, "y": [1, 2, None, 3, 4]}]])
        >>> ak.to_list(ak.full_like(array, 12.3))
        [[{"x": 12.3, "y": []},
          {"x": 12.3, "y": [12]},
          {"x": 12.3, "y": [12, 12]}],
         [],
         [{"x": 12.3, "y": [12, 12, None, 12]},
          True,
          True,
          True,
          {"x": 12.3, "y": [12, 12, None, 12, 12]}]]

    The `"x"` values get filled in with `12.3` because they retain their type
    (`float64`) and the `"y"` list items get filled in with `12` because they
    retain their type (`int64`). Booleans get filled with True because `12.3`
    is not zero. Missing values remain in the same positions as in the original
    `array`. (To fill them in, use #ak.fill_none.)

    See also #ak.zeros_like and #ak.ones_like.

    (There is no equivalent of NumPy's `np.empty_like` because Awkward Arrays
    are immutable.)
    """
    if dtype is not None:
        # In the case of strings and byte strings,
        # converting the fill avoids a ValueError.
        dtype = np.dtype(dtype)
        nplike = ak.nplike.of(array)
        fill_value = nplike.array([fill_value], dtype=dtype)[0]
        # Also, if the fill_value cannot be converted to the dtype
        # this should throw a clear, early, error.
        if dtype == np.dtype(np.bool_):
            # then for bools, only 0 and 1 give correct string behavior
            fill_value = fill_value.view(np.uint8)

    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=True, allow_other=False
    )

    def action(layout, **kwargs):
        if layout.parameter("__array__") == "bytestring" and fill_value is _ZEROS:
            nplike = ak.nplike.of(layout)
            asbytes = nplike.frombuffer(b"", dtype=np.uint8)
            return ak._v2.contents.ListArray(
                ak._v2.index.Index64(nplike.zeros(len(layout), dtype=np.int64)),
                ak._v2.index.Index64(nplike.zeros(len(layout), dtype=np.int64)),
                ak._v2.contents.NumpyArray(asbytes, parameters={"__array__": "byte"}),
                parameters={"__array__": "bytestring"},
            )

        elif layout.parameter("__array__") == "bytestring":
            nplike = ak.nplike.of(layout)
            if isinstance(fill_value, bytes):
                asbytes = fill_value
            else:
                asbytes = str(fill_value).encode("utf-8", "surrogateescape")
            asbytes = nplike.frombuffer(asbytes, dtype=np.uint8)

            return ak._v2.contents.ListArray(
                ak._v2.index.Index64(nplike.zeros(len(layout), dtype=np.int64)),
                ak._v2.index.Index64(
                    nplike.full(len(layout), len(asbytes), dtype=np.int64)
                ),
                ak._v2.contents.NumpyArray(asbytes, parameters={"__array__": "byte"}),
                parameters={"__array__": "bytestring"},
            )

        elif layout.parameter("__array__") == "string" and fill_value is _ZEROS:
            nplike = ak.nplike.of(layout)
            asbytes = nplike.frombuffer(b"", dtype=np.uint8)
            return ak._v2.contents.ListArray(
                ak._v2.index.Index64(nplike.zeros(len(layout), dtype=np.int64)),
                ak._v2.index.Index64(nplike.zeros(len(layout), dtype=np.int64)),
                ak._v2.contents.NumpyArray(asbytes, parameters={"__array__": "char"}),
                parameters={"__array__": "string"},
            )

        elif layout.parameter("__array__") == "string":
            nplike = ak.nplike.of(layout)
            asstr = str(fill_value).encode("utf-8", "surrogateescape")
            asbytes = nplike.frombuffer(asstr, dtype=np.uint8)
            return ak._v2.contents.ListArray(
                ak._v2.index.Index64(nplike.zeros(len(layout), dtype=np.int64)),
                ak._v2.index.Index64(
                    nplike.full(len(layout), len(asbytes), dtype=np.int64)
                ),
                ak._v2.contents.NumpyArray(asbytes, parameters={"__array__": "char"}),
                parameters={"__array__": "string"},
            )

        elif isinstance(layout, ak._v2.contents.NumpyArray):
            nplike = ak.nplike.of(layout)
            original = nplike.asarray(layout)

            if fill_value == 0 or fill_value is _ZEROS:
                return ak._v2.contents.NumpyArray(
                    nplike.zeros_like(original),
                    layout.identifier,
                    layout.parameters,
                )
            elif fill_value == 1:
                return ak._v2.contents.NumpyArray(
                    nplike.ones_like(original),
                    layout.identifier,
                    layout.parameters,
                )
            else:
                return ak._v2.contents.NumpyArray(
                    nplike.full_like(original, fill_value),
                    layout.identifier,
                    layout.parameters,
                )
        else:
            return None

    out = layout.recursively_apply(action)
    if dtype is not None:
        out = ak._v2.operations.structure.strings_astype(
            out, dtype, highlevel, behavior
        )
        out = ak._v2.operations.structure.values_astype(out, dtype, highlevel, behavior)
        return out
    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def singletons(array, highlevel=True, behavior=None):
    raise NotImplementedError


#     """
#     Args:
#         array: Data to wrap in lists of length 1 if present and length 0
#             if missing (None).
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.
#         behavior (None or dict): Custom #ak.behavior for the output array, if
#             high-level.

#     Returns a singleton list (length 1) wrapping each non-missing value and
#     an empty list (length 0) in place of each missing value.

#     For example,

#         >>> array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
#         >>> print(ak.singletons(array))
#         [[1.1], [2.2], [], [3.3], [], [], [4.4], [5.5]]

#     See #ak.firsts to invert this function.
#     """

#     def getfunction(layout):
#         nplike = ak.nplike.of(layout)

#         if isinstance(layout, ak._util.optiontypes):
#             nulls = nplike.asarray(layout.bytemask()).view(np.bool_)
#             offsets = nplike.ones(len(layout) + 1, dtype=np.int64)
#             offsets[0] = 0
#             offsets[1:][nulls] = 0
#             nplike.cumsum(offsets, out=offsets)
#             return lambda: ak.layout.ListOffsetArray64(
#                 ak.layout.Index64(offsets), layout.project()
#             )

#         elif isinstance(layout, ak.layout.IndexedArray32) and isinstance(
#             layout.content, (ak.layout.EmptyArray, ak.layout.NumpyArray)
#         ):
#             return getfunction(
#                 ak.layout.IndexedOptionArray32(
#                     layout.index,
#                     layout.content,
#                     layout.identities,
#                     layout.parameters,
#                 )
#             )

#         elif isinstance(layout, ak.layout.IndexedArrayU32) and isinstance(
#             layout.content, (ak.layout.EmptyArray, ak.layout.NumpyArray)
#         ):
#             return getfunction(
#                 ak.layout.IndexedOptionArray64(
#                     ak.layout.Index64(layout.index),
#                     layout.content,
#                     layout.identities,
#                     layout.parameters,
#                 )
#             )

#         elif isinstance(layout, ak.layout.IndexedArray64) and isinstance(
#             layout.content, (ak.layout.EmptyArray, ak.layout.NumpyArray)
#         ):
#             return getfunction(
#                 ak.layout.IndexedOptionArray64(
#                     layout.index,
#                     layout.content,
#                     layout.identities,
#                     layout.parameters,
#                 )
#             )

#         elif isinstance(layout, ak.layout.EmptyArray):
#             return getfunction(ak.layout.UnmaskedArray(layout.toNumpyArray()))

#         elif isinstance(layout, ak.layout.NumpyArray):
#             return getfunction(ak.layout.UnmaskedArray(layout))

#         else:
#             return None

#     layout = ak._v2.operations.convert.to_layout(array)
#     out = ak._v2._util.recursively_apply(layout, getfunction, pass_depth=False)

#     return ak._v2._util.maybe_wrap_like(out, array, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def argcombinations(
    array,
    n,
    replacement=False,
    axis=1,
    fields=None,
    parameters=None,
    with_name=None,
    highlevel=True,
    behavior=None,
):

    """
    Args:
        array: Array from which to choose `n` items without replacement.
        n (int): The number of items to choose from each list: `2` chooses
            unique pairs, `3` chooses unique triples, etc.
        replacement (bool): If True, combinations that include the same
            item more than once are allowed; otherwise each item in a
            combinations is strictly unique.
        axis (int): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        fields (None or list of str): If None, the pairs/triples/etc. are
            tuples with unnamed fields; otherwise, these `fields` name the
            fields. The number of `fields` must be equal to `n`.
        parameters (None or dict): Parameters for the new
            #ak.layout.RecordArray node that is created by this operation.
        with_name (None or str): Assigns a `"__record__"` name to the new
            #ak.layout.RecordArray node that is created by this operation
            (overriding `parameters`, if necessary).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.

    Computes a Cartesian product (i.e. cross product) of `array` with itself
    that is restricted to combinations sampled without replacement,
    like #ak.combinations, but returning integer indexes for
    #ak.Array.__getitem__.

    The motivation and uses of this function are similar to those of
    #ak.argcartesian. See #ak.combinations and #ak.argcartesian for a more
    complete description.
    """
    if parameters is None:
        parameters = {}
    else:
        parameters = dict(parameters)
    if with_name is not None:
        parameters["__record__"] = with_name

    if axis < 0:
        raise ValueError("the 'axis' for argcombinations must be non-negative")
    else:
        layout = ak._v2.operations.convert.to_layout(
            array, allow_record=False, allow_other=False
        ).localindex(axis)
        out = layout.combinations(
            n, replacement=replacement, axis=axis, fields=fields, parameters=parameters
        )
        return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

# v2: drop this file, but the from_json functionality must be available somehow.
# (uproot_issue_90 will not be ported; it will be replaced by AwkwardForth.)

from awkward._ext import fromjson
from awkward._ext import uproot_issue_90


__all__ = ["fromjson", "uproot_issue_90"]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
import json

np = ak.nplike.NumpyMetadata.instance()


class NDJSONDecoder(json.JSONDecoder):
    """
    JSON decoder class to implement 'ndjson' specs:
    see https://github.com/ndjson/ndjson-spec

    - accept newline as line delimiter '\n' (0x0A) as well as carriage return and newline '\r\n' (0x0D0A)
    - ignore empty lines '\n\n'
    - raise an error if JSON is nor parsable
    """

    def decode(self, s, *args, **kwargs):
        lines = s.splitlines()
        non_empty_lines = [line for line in lines if line.strip() != ""]
        lines = ",".join(non_empty_lines)
        text = f"[{lines}]"
        return super().decode(text, *args, **kwargs)


def from_json_file(
    source,
    nan_string=None,
    infinity_string=None,
    minus_infinity_string=None,
    complex_record_fields=None,
    highlevel=False,
    behavior=None,
    buffersize=65536,
):
    """
    Args:
        source (str): A filename with a JSON-formatted content to convert into an array.
        nan_string (None or str): If not None, strings with this value will be
            interpreted as floating-point NaN values.
        infinity_string (None or str): If not None, strings with this value will
            be interpreted as floating-point positive infinity values.
        minus_infinity_string (None or str): If not None, strings with this value
            will be interpreted as floating-point negative infinity values.
        complex_record_fields (None or (str, str)): If not None, defines a pair of
            field names to interpret records as complex numbers.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        buffersize (int): Size (in bytes) of the buffer used by the JSON
            parser.

    Converts content of a JSON file into an Awkward Array.

    See also #ak.from_json_schema and #ak.to_json.
    """
    if complex_record_fields is None:
        complex_real_string = None
        complex_imag_string = None
    elif (
        isinstance(complex_record_fields, tuple)
        and len(complex_record_fields) == 2
        and isinstance(complex_record_fields[0], str)
        and isinstance(complex_record_fields[1], str)
    ):
        complex_real_string, complex_imag_string = complex_record_fields

    with open(source, encoding="utf-8") as f:
        data = json.load(f, cls=NDJSONDecoder)

    layout = ak._v2.operations.convert.from_iter(data).layout

    def record_to_complex(node, **kwargs):
        if isinstance(node, ak._v2.contents.RecordArray):
            keys = node.fields
            if complex_record_fields[0] in keys and complex_record_fields[1] in keys:
                real = node[complex_record_fields[0]]
                imag = node[complex_record_fields[1]]
                if (
                    isinstance(real, ak._v2.contents.NumpyArray)
                    and len(real.shape) == 1
                    and isinstance(imag, ak._v2.contents.NumpyArray)
                    and len(imag.shape) == 1
                ):
                    return ak._v2.contents.NumpyArray(
                        node._nplike.asarray(real) + node._nplike.asarray(imag) * 1j
                    )
                else:
                    raise ValueError("Complex number fields must be numbers")
                return ak._v2.contents.NumpyArray(real + imag * 1j)

    layout = (
        layout
        if complex_imag_string is None
        else layout.recursively_apply(record_to_complex)
    )

    nonfinite_dict = {}
    if nan_string is not None:
        nonfinite_dict[nan_string] = np.nan
    if infinity_string is not None:
        nonfinite_dict[infinity_string] = np.inf
    if minus_infinity_string is not None:
        nonfinite_dict[minus_infinity_string] = -np.inf

    def string_to_nonfinite(node, **kwargs):
        if isinstance(node, ak._v2.contents.ListOffsetArray):
            if node.parameter("__array__") == "string":
                return node._awkward_strings_to_nonfinite(nonfinite_dict)

    layout = layout.recursively_apply(string_to_nonfinite) if nonfinite_dict else layout

    layout = (
        layout.content
        if isinstance(layout, ak._v2.contents.ListOffsetArray)
        else layout
    )

    return ak._v2._util.wrap(layout, behavior, highlevel)
